/*==============================================================================
;	   PIC16F876A  LCD 16x2
;	   20 MHz
;=============================================================================*/
#define	PICDEV    0         // 0  PIC16F876. 1  PIC16F886
#if  PICDEV
// PIC16F886 Configuration Bit Settings
// CONFIG1
#pragma config FOSC = HS        // Oscillator Selection bits (HS oscillator: High-speed crystal/resonator on RA6/OSC2/CLKOUT and RA7/OSC1/CLKIN)
#pragma config WDTE = OFF       // Watchdog Timer Enable bit (WDT disabled and can be enabled by SWDTEN bit of the WDTCON register)
#pragma config PWRTE = ON       // Power-up Timer Enable bit (PWRT enabled)
#pragma config MCLRE = ON       // RE3/MCLR pin function select bit (RE3/MCLR pin function is MCLR)
#pragma config CP = OFF         // Code Protection bit (Program memory code protection is disabled)
#pragma config CPD = OFF        // Data Code Protection bit (Data memory code protection is disabled)
#pragma config BOREN = OFF      // Brown Out Reset Selection bits (BOR disabled)
#pragma config IESO = ON        // Internal External Switchover bit (Internal/External Switchover mode is enabled)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enabled bit (Fail-Safe Clock Monitor is enabled)
#pragma config LVP = OFF        // Low Voltage Programming Enable bit (RB3 pin has digital I/O, HV on MCLR must be used for programming)
// CONFIG2
#pragma config BOR4V = BOR40V   // Brown-out Reset Selection bit (Brown-out Reset set to 4.0V)
#pragma config WRT = OFF        // Flash Program Memory Self Write Enable bits (Write protection off)
#else
// PIC16F876A Configuration Bit Settings
#pragma config FOSC = HS    //     
#pragma config WDTE = OFF   //   
#pragma config PWRTE = ON   //    
#pragma config BOREN = OFF  //      
#pragma config LVP = OFF    //   , RB3-I/O 
#pragma config CPD = OFF    //    
#pragma config WRT = OFF    //    FLASH  EECON
#pragma config CP = OFF     //    
#endif

#include <pic.h>
#include <stdio.h>
#include <stdlib.h>
#include "lcd.h"

#define	Count0    RC2       //   Fr 
#define	Count1    RC1       //   Fx
#define	GATE_OUT  RC3       //  
#define	RESD      RC4       //  
#define	GATE_IN   RC5       //   5 
//-----------  ------------------------------------------------------//
void init();                // 
void interrupt tmr2(void);  //   Timmer2  2ms
char readAdc(void);         //   
void usign(char k, char l); //    
void ust(void);             //   
void t0count(void);         //  Timer0
void t1count(void);         //  Timer1
void lcd_0Hz(void);         //  0 Hz
void dir_acc(void);         //    
void rev_acc(void);         //    
unsigned char schet(unsigned long n); //    

typedef struct {
    unsigned Meas   :1;     // 
    unsigned d_10   :1;     //  0.5 
    unsigned Flst   :1;     //   splash screen
    unsigned Flow   :1;     //    F < 0.1 
    unsigned Fdiv   :1;     //
    unsigned Flag6  :3;
} Fbits_t;
//-----------  -----------------------------------------------------//
Fbits_t Flags = 0;
unsigned char l,k;          //   CGRAM, 
unsigned char df2;          //   250 ( 2)
unsigned char d10;          //   10  5
unsigned char UF;           //  , 
unsigned int x;
unsigned int Fx,Fr;
unsigned long M,N,z;
double Fi;
char buffer[11];

const char arr_cgram[4][8] = { //'',|,||,|||
0x10,0x10,0x15,0,0,0,0,0,0x10,0x10,0x15,0,0x10,0x10,0x10,0x10,
0x10,0x10,0x15,0,0x14,0x14,0x14,0x14,0x10,0x10,0x15,0,0x15,0x15,0x15,0x15};
const char arr_sim[2][17] = {"Frequency  meter","radiokot.ru 2017"};
const char arr_s0[] = {"F=          0 Hz"}; 
const char arr_div[2][4] = {":64","1:1"}; 

void main (void) {
init();                     // 
do {                        //  2 
    if (Flags.d_10) { Flags.d_10 = 0; if (d10++ > 3) Flags.Flst = 1; }
}   while(!Flags.Flst);
lcd_clear();
lcd_0Hz();                  // 0 Hz
ust();                      //   
while(1){                   //   
if (Flags.d_10) {           //   0.5 
Flags.d_10 = 0;
//-----------------------  5  ------------------------------------//
if (Flags.Meas) {
    if (d10++ >= 9) {       //  5 
        d10 = 0;
        if (!GATE_OUT) Flags.Flow = 1;//    F=0
        Flags.Meas=0; }
    Flags.Fdiv = RA2;       //   ?
    lcd_gotoxy(1,3);
    lcd_ram((char*)arr_div[Flags.Fdiv],3);//   
    UF = d10;
    usign(0,3);             //   
    UF = readAdc();         //   
    usign(6,10);            //     
}
//----------------------    ------------------------------//
else {
    if (Flags.Flow) {       //     F=0
        lcd_0Hz();          // 0 Hz 
        ust(); }             //   
    else {
    if (!GATE_OUT) {        //     
        if ((Fx > 0x0A) | !Flags.Fdiv) { //  2.1   
        dir_acc();          //    
        }
        else {
        rev_acc();          //    
        }
        ust(); 
    }
    else {                  //   ,  
        if (d10++ > 2) {    // 
        lcd_0Hz();          // 0 Hz 
        ust();
}}}}}}}
//---------------------    -----------------------------//

unsigned char schet(unsigned long n) {
UF = 1;
while ( n = n / 10 ) UF++;
return UF;
}

void rev_acc(void) {        //    
t1count();
N = Fx;
N = N << 20;
z = x;
z = z <<4;
N = N + z + 14 - UF ;
t0count();
M = Fr;
M = M << 16;
z = x;
z = z <<8;
M = M + z +255 - UF ;
Fi = ( 20000000.0 * (double)N ) / (double)M ;
z = (unsigned long) Fi;
k = 8 - schet(z);
lcd_0Hz();
lcd_gotoxy(0,4);
sprintf(buffer,"%8.*f",k,Fi);
lcd_puts(buffer);
}

void dir_acc(void) {        //    
t1count();
N = Fx;
N = N << 20;
z = x;
z = z <<4;
z = z + N + 14 - UF ;
lcd_0Hz(); 
if (!Flags.Fdiv) {          //   
    z = z << 6;
    N = z / 5000;
    x = z % 5000;
    x = x * 2 / 100;
    sprintf(buffer,"%7ld.%02d k",N,x); 
    }
else {
    N = z / 5;
    x = (z % 5)*2;
    sprintf(buffer,"%9ld.%1d",N,x);
    }
lcd_gotoxy(0,2);
lcd_puts(buffer);
}

void lcd_0Hz(void) {
lcd_gotoxy(0,0);
lcd_ram((char*)arr_s0,16);
}

void t0count(void) {        //  Timer0
UF = 0;
x = TMR0; 
k = TMR0 & 0x01;
if (k) {
    for (i=0;i<255;i++) { 
        Count0=0;Delayus(2);Count0=1;
        k = TMR0 & 0x01; 
        if (k) ++UF;
        else break; }
    }
else {
    for (i=0;i<255;i++) { 
        Count0=0;Delayus(2);Count0=1;
        k = TMR0 & 0x01; 
        if (!k) ++UF;
        else break; }
    }
}

void t1count(void) {        //  Timer1
UF = 0;
x = TMR1L; 
x += TMR1H << 8;
k = TMR1L & 0x01;
if (k) {
    for (i=0;i<15;i++) { 
        Count1=0;Delayus(2);Count1=1;
        k = TMR1L & 0x01;
        if (k) ++UF;
        else break; }
    }
else {
    for (i=0;i<15;i++) { 
        Count1=0;Delayus(2);Count1=1;
        k = TMR1L & 0x01;
        if (!k) ++UF;
        else break; }
    }
}

void init() {       // 
INTCON=0;           //   
TRISC=0b00001001;   // PORTC 1,2,4-7  0,3 
TRISB=0;            // PORTB 0-7 
TRISA=0xFF;         // PORTA 0-5 
OPTION_REG=0xE7;    // TMR0  RA4,  , .TMR0- 256, .-.
#if PICDEV
ANSEL=0b00001001;
ADCON1=0x10;
#else
ADCON1=0x05;        //  RA3-VREF+, AN0,AN1-, AN2,4,5-. .
#endif
PR2=249;            // TMR2 :250 500 Hz
TMR2IE=1;           //    TMR2
T2CON=0b01001101;   // . TMR2 .  1:10, 1:4
T1CON=0b00110111;   // . TMR1 .  1:8, -, .  RC0
PEIE=1;             //     
lcd_init();         //  
lcd_clear();        //  
for(l=0;l<4;l++) lcd_cgram((char*)arr_cgram[l], l);  //   CGRAM
lcd_gotoxy(0,0);
lcd_ram((char*)arr_sim[0],16); // Frecuency meter
lcd_gotoxy(1,0);
lcd_ram((char*)arr_sim[1],16); // RadioKot 2017
df2 = 1; d10 = 0;   //  
GIE = 1;            //   
 }

void interrupt tmr2(void) { //    2 500
if(!(--df2)) {
    if (Flags.Meas) {
        asm("NOP");
        asm("NOP");
        GATE_IN = 1;
    }
    else { 
        asm("NOP");
        GATE_IN = 0;
    }
//    GATE_IN=Flags.Meas;    //     
    df2=250;
    Flags.d_10=1;
    }
if(TMR1IF) { Fx++; TMR1IF=0; }  //   TMR1
if(TMR0IF) { Fr++; TMR0IF=0; }  //   TMR0
TMR2IF = 0;         //   
}

unsigned char readAdc() {   //   
ADCON0=0b10000001;	//   ,   AN0 
Delayus(20);        //     20 
GO_DONE =1;         //  .  .
while(GO_DONE == 1);//   - 
return(ADRESH >> 3);// 
}

void usign(char k,char l){   //      
lcd_gotoxy(1,k);
for(i=0;i<l;i++) lcd_putch(0);
lcd_gotoxy(1,k);
for(i=0;i<l;i++) { 
    if (UF>3) { lcd_putch(3); UF=UF-3; }
    else { lcd_putch(UF); break; }
    }
}

void ust(void) {      //   
Count0 = 1;           //   20 
Count1 = 1;           //   Fx
TMR1H = 0;            //   TMR1
TMR1L = 0;            //  
Fx = 0;               //   Fx
TMR0 = 0;             //   TMR0
Fr = 0;               //   20 
RESD = 0; Delayus(2); RESD = 1; //   
d10 = 0;              //   10  5
Flags.Meas = 1;       // 
Flags.Flow = 0;       //    F  0.2
df2 = 1;              //     250 ( 2)
}